"""
Copyright (c) 2015, 2020 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

#
# This file is generated from template: rel-eng/version_init.template
# DO NOT MODIFY THIS FILE
# If changes are necessary, edit ./rel-eng/version_init.template
#

from __future__ import absolute_import
from .parser import DockerfileParser  # noqa: F401

__version__ = "2.0.0"



# -*- coding: utf-8 -*-
"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

import io

import pytest


from dockerfile_parse import DockerfileParser


@pytest.fixture(
    params=[(use_fileobj, cache_content)
            for use_fileobj in [True, False]
            for cache_content in [True, False]])
def dfparser(tmpdir, request):
    """

    :param tmpdir: already existing fixture defined in pytest
    :param request: parameter, cache_content arg to DockerfileParser
    :return: DockerfileParser instance
    """

    use_fileobj, cache_content = request.param
    if use_fileobj:
        fileobj = io.BytesIO()
        return DockerfileParser(fileobj=fileobj, cache_content=cache_content)
    else:
        tmpdir_path = str(tmpdir.realpath())
        return DockerfileParser(path=tmpdir_path, cache_content=cache_content)


@pytest.fixture(params=['LABEL', 'ENV', 'ARG'])
def instruction(request):
    """
    Parametrized fixture which enables to run a test once for each instruction in params
    """
    return request.param



# -*- coding: utf-8 -*-
"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from io import StringIO


def b2u(string):
    """ bytes to unicode """
    if isinstance(string, bytes):
        return string.decode('utf-8')
    return string


def u2b(string):
    """ unicode to bytes"""
    if isinstance(string, str):
        return string.encode('utf-8')
    return string


class WordSplitter(object):
    """
    Split string into words, substituting environment variables if provided

    Methods defined here:

    dequote()
        Returns the string with escaped and quotes consumed

    split(maxsplit=None, dequote=True)
        Returns an iterable of words, split at whitespace
    """

    SQUOTE = "'"
    DQUOTE = '"'

    def __init__(self, s, args=None, envs=None):
        """
        :param s: str, string to process
        :param args: dict, build arguments to use; if None, do not
            attempt substitution
        :param envs: dict, environment variables to use; if None, do not
            attempt substitution
        """
        self.stream = StringIO(s)
        self.args = args
        self.envs = envs

        # Initial state
        self.quotes = None  # the quoting character in force, or None
        self.escaped = False

    def _update_quoting_state(self, ch):
        """
        Update self.quotes and self.escaped

        :param ch: str, current character
        :return: ch if it was not used to update quoting state, else ''
        """

        # Set whether the next character is escaped
        # Unquoted:
        #   a backslash escapes the next character
        # Double-quoted:
        #   a backslash escapes the next character only if it is a double-quote
        # Single-quoted:
        #   a backslash is not special
        is_escaped = self.escaped
        self.escaped = (not self.escaped and
                        ch == '\\' and
                        self.quotes != self.SQUOTE)
        if self.escaped:
            return ''

        if is_escaped:
            if self.quotes == self.DQUOTE:
                if ch == '"':
                    return ch
                return "{0}{1}".format('\\', ch)

            return ch

        if self.quotes is None:
            if ch in (self.SQUOTE, self.DQUOTE):
                self.quotes = ch
                return ''

        elif self.quotes == ch:
            self.quotes = None
            return ''

        return ch

    def dequote(self):
        return ''.join(self.split(maxsplit=0))

    def split(self, maxsplit=None, dequote=True):
        """
        Generator for the words of the string

        :param maxsplit: perform at most maxsplit splits;
            if None, do not limit the number of splits
        :param dequote: remove quotes and escape characters once consumed
        """

        class Word(object):
            """
            A None-or-str object which can always be appended to.
            Similar to a defaultdict but with only a single value.
            """

            def __init__(self):
                self.value = None

            @property
            def valid(self):
                return self.value is not None

            def append(self, s):
                if self.value is None:
                    self.value = s
                else:
                    self.value += s

        num_splits = 0
        word = Word()
        while True:
            ch = self.stream.read(1)
            if not ch:
                # EOF
                if word.valid:
                    yield word.value

                return

            if (not self.escaped and
                    (self.envs is not None or self.args is not None) and
                    ch == '$' and
                    self.quotes != self.SQUOTE):
                while True:
                    # Substitute environment variable
                    braced = False
                    varname = ''
                    while True:
                        ch = self.stream.read(1)
                        if varname == '' and ch == '{':
                            braced = True
                            continue

                        if not ch:
                            # EOF
                            break

                        if braced and ch == '}':
                            break

                        if not ch.isalnum() and ch != '_':
                            break

                        varname += ch

                    if self.envs is not None and varname in self.envs:
                        word.append(self.envs[varname])
                    elif self.args is not None and varname in self.args:
                        word.append(self.args[varname])

                    # Check whether there is another envvar
                    if ch != '$':
                        break

                if braced and ch == '}':
                    continue

                # ch now holds the next character

            # Figure out what our quoting/escaping state will be
            # after this character
            is_escaped = self.escaped
            ch_unless_consumed = self._update_quoting_state(ch)

            if dequote:
                # If we just processed a quote or escape character,
                # and were asked to dequote the string, consume it now
                ch = ch_unless_consumed

            # If word-splitting has been requested, check whether we are
            # at a whitespace character
            may_split = maxsplit != 0 and (maxsplit is None or
                                           num_splits < maxsplit)
            at_split = may_split and (self.quotes is None and
                                      not is_escaped and
                                      ch.isspace())
            if at_split:
                # It is time to yield a word
                if word.valid:
                    num_splits += 1
                    yield word.value

                word = Word()
            else:
                word.append(ch)


def extract_key_values(env_replace, args, envs, instruction_value):
    words = list(WordSplitter(instruction_value).split(dequote=False))
    key_val_list = []

    def substitute_vars(val):
        kwargs = {}
        if env_replace:
            kwargs['args'] = args
            kwargs['envs'] = envs

        return WordSplitter(val, **kwargs).dequote()

    if '=' not in words[0]:
        # This form is:
        #   LABEL/ENV name value
        # The first word is the name, remainder are the value.
        key_val = [substitute_vars(x) for x in instruction_value.split(None, 1)]
        key = key_val[0]
        try:
            val = key_val[1]
        except IndexError:
            val = ''

        key_val_list.append((key, val))
    else:
        # This form is:
        #   LABEL/ENV "name"="value" ["name"="value"...]
        # Each word is a key=value pair.
        for k_v in words:
            if '=' not in k_v:
                raise ValueError('Syntax error - can\'t find = in "{word}". '
                                 'Must be of the form: name=value'
                                 .format(word=k_v))
            key, val = [substitute_vars(x) for x in k_v.split('=', 1)]
            key_val_list.append((key, val))

    return key_val_list


def get_key_val_dictionary(instruction_value, env_replace=False, args=None, envs=None):
    args = args or {}
    envs = envs or {}
    return dict(extract_key_values(instruction_value=instruction_value,
                                   env_replace=env_replace,
                                   args=args, envs=envs))


class Context(object):
    def __init__(self, args=None, envs=None, labels=None,
                 line_args=None, line_envs=None, line_labels=None):
        """
        Class representing current state of build arguments, environment variables and labels.

        :param args: dict with arguments valid for this line
            (all variables defined to this line)
        :param envs: dict with variables valid for this line
            (all variables defined to this line)
        :param labels: dict with labels valid for this line
            (all labels defined to this line)
        :param line_args: dict with arguments defined on this line
        :param line_envs: dict with variables defined on this line
        :param line_labels: dict with labels defined on this line
        """
        self.args = args or {}
        self.envs = envs or {}
        self.labels = labels or {}
        self.line_args = line_args or {}
        self.line_envs = line_envs or {}
        self.line_labels = line_labels or {}

    def set_line_value(self, context_type, value):
        """
        Set value defined on this line ('line_args'/'line_envs'/'line_labels')
        and update 'args'/'envs'/'labels'.

        :param context_type: "ARG" or "ENV" or "LABEL"
        :param value: new value for this line
        """
        if context_type.upper() == "ARG":
            self.line_args = value
            self.args.update(value)
        elif context_type.upper() == "ENV":
            self.line_envs = value
            self.envs.update(value)
        elif context_type.upper() == "LABEL":
            self.line_labels = value
            self.labels.update(value)
        else:
            raise ValueError("Unexpected context type: " + context_type)

    def get_line_value(self, context_type):
        """
        Get the values defined on this line.

        :param context_type: "ARG" or "ENV" or "LABEL"
        :return: values of given type defined on this line
        """
        if context_type.upper() == "ARG":
            return self.line_args
        if context_type.upper() == "ENV":
            return self.line_envs
        if context_type.upper() == "LABEL":
            return self.line_labels
        raise ValueError("Unexpected context type: " + context_type)

    def get_values(self, context_type):
        """
        Get the values valid on this line.

        :param context_type: "ARG" or "ENV" or "LABEL"
        :return: values of given type valid on this line
        """
        if context_type.upper() == "ARG":
            return self.args
        if context_type.upper() == "ENV":
            return self.envs
        if context_type.upper() == "LABEL":
            return self.labels
        raise ValueError("Unexpected context type: " + context_type)



# -*- coding: utf-8 -*-
"""
Copyright (c) 2015, 2019 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

import inspect
import io
import json
import os
import pytest
import re
import sys
from textwrap import dedent

from dockerfile_parse import DockerfileParser
from dockerfile_parse.parser import image_from
from dockerfile_parse.constants import COMMENT_INSTRUCTION
from dockerfile_parse.util import b2u, u2b, Context
from tests.fixtures import dfparser, instruction

NON_ASCII = "žluťoučký"
# flake8 does not understand fixtures:
dfparser = dfparser  # pylint: disable=self-assigning-variable
instruction = instruction  # pylint: disable=self-assigning-variable


class TestDockerfileParser(object):
    def test_all_versions_match(self):
        def read_version(fp, regex):
            with open(fp, "r") as fd:
                content = fd.read()
                found = re.findall(regex, content)
                if len(found) == 1:
                    return found[0]
                else:
                    raise Exception("Version not found!")

        import dockerfile_parse
        from dockerfile_parse import __version__ as module_version
        fp = inspect.getfile(dockerfile_parse)
        project_dir = os.path.dirname(os.path.dirname(fp))
        specfile = os.path.join(project_dir, "python-dockerfile-parse.spec")
        setup_py = os.path.join(project_dir, "setup.py")
        spec_version = read_version(specfile, r"\nVersion:\s*(.+?)\s*\n")
        setup_py_version = read_version(setup_py, r"version=['\"](.+)['\"]")
        assert spec_version == module_version
        assert setup_py_version == module_version

    def test_util_b2u(self):
        assert isinstance(b2u('string'), str)
        assert isinstance(b2u(b'byte'), str)

    def test_util_u2b(self):
        assert isinstance(u2b('string'), bytes)
        assert isinstance(u2b(b'byte'), bytes)

    def test_util_context_exceptions(self):
        context = Context()
        with pytest.raises(ValueError):
            context.get_values('FOO')
        with pytest.raises(ValueError):
            context.get_line_value('FOO')
        with pytest.raises(ValueError):
            context.set_line_value('FOO', {})

    def test_dockerfileparser(self, dfparser, tmpdir):
        df_content = dedent("""\
            FROM fedora
            LABEL label={0}""".format(NON_ASCII))
        df_lines = ["FROM fedora\n", "LABEL label={0}".format(NON_ASCII)]

        dfparser.content = ""
        dfparser.content = df_content
        assert dfparser.content == df_content
        assert dfparser.lines == df_lines
        assert [isinstance(line, str) for line in dfparser.lines]

        dfparser.content = ""
        dfparser.lines = df_lines
        assert dfparser.content == df_content
        assert dfparser.lines == df_lines
        assert [isinstance(line, str) for line in dfparser.lines]

        dockerfile = os.path.join(str(tmpdir), 'Dockerfile')
        with open(dockerfile, 'wb') as fp:
            fp.write(df_content.encode('utf-8'))
        dfparser = DockerfileParser(dockerfile)
        assert dfparser.content == df_content
        assert dfparser.lines == df_lines
        assert [isinstance(line, str) for line in dfparser.lines]

    def test_dockerfileparser_exceptions(self, tmpdir):
        df_content = dedent("""\
            FROM fedora
            LABEL label={0}""".format(NON_ASCII))
        df_lines = ["FROM fedora\n", "LABEL label={0}".format(NON_ASCII)]

        dfp = DockerfileParser(os.path.join(str(tmpdir), 'no-directory'))
        with pytest.raises(IOError):
            assert dfp.content
        with pytest.raises(IOError):
            dfp.content = df_content
        with pytest.raises(IOError):
            assert dfp.lines
        with pytest.raises(IOError):
            dfp.lines = df_lines

    def test_internal_exceptions(self, tmpdir):
        dfp = DockerfileParser(str(tmpdir))
        with pytest.raises(ValueError):
            dfp._instruction_getter('FOO', env_replace=True)
        with pytest.raises(ValueError):
            dfp._instructions_setter('FOO', {})
        with pytest.raises(ValueError):
            dfp._modify_instruction_label_env('FOO', 'key', 'value')

    def test_constructor_cache(self, tmpdir):
        tmpdir_path = str(tmpdir.realpath())
        df1 = DockerfileParser(tmpdir_path)
        df1.lines = ["From fedora:latest\n", "LABEL a b\n"]

        df2 = DockerfileParser(tmpdir_path, True)
        assert df2.cached_content

    def test_dockerfile_structure(self, dfparser):
        dfparser.lines = ["# comment\n",                # single-line comment
                          " From  \\\n",                # mixed-case
                          "   base\n",                  # extra ws, continuation line
                          " #    another   comment\n",  # extra ws
                          " label  foo  \\\n",          # extra ws
                          "# interrupt LABEL\n",        # comment interrupting multi-line LABEL
                          "    bar  \n",                # extra ws, instruction continuation
                          "USER  {0}\n".format(NON_ASCII),
                          "# comment \\\n",             # extra ws
                          "# with \\ \n",               # extra ws with a space
                          "# backslashes \\\\ \n",      # two backslashes
                          "#no space after hash\n",
                          "# comment # with hash inside\n",
                          "RUN command1\n",
                          "RUN command2 && \\\n",
                          "    command3\n",
                          "RUN command4 && \\\n",
                          "# interrupt RUN\n",          # comment interrupting multi-line RUN
                          "    command5\n",
                          ]

        assert dfparser.structure == [
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 0,
                                       'endline': 0,
                                       'content': '# comment\n',
                                       'value': 'comment'},
                                      {'instruction': 'FROM',
                                       'startline': 1,
                                       'endline': 2,
                                       'content': ' From  \\\n   base\n',
                                       'value': 'base'},
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 3,
                                       'endline': 3,
                                       'content': ' #    another   comment\n',
                                       'value': 'another   comment'},
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 5,
                                       'endline': 5,
                                       'content': '# interrupt LABEL\n',
                                       'value': 'interrupt LABEL'},
                                      {'instruction': 'LABEL',
                                       'startline': 4,
                                       'endline': 6,
                                       'content': ' label  foo  \\\n    bar  \n',
                                       'value': 'foo      bar'},
                                      {'instruction': 'USER',
                                       'startline': 7,
                                       'endline': 7,
                                       'content': 'USER  {0}\n'.format(NON_ASCII),
                                       'value': '{0}'.format(NON_ASCII)},
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 8,
                                       'endline': 8,
                                       'content': '# comment \\\n',
                                       'value': 'comment \\'},
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 9,
                                       'endline': 9,
                                       'content': '# with \\ \n',
                                       'value': 'with \\ '},
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 10,
                                       'endline': 10,
                                       'content': '# backslashes \\\\ \n',
                                       'value': 'backslashes \\\\ '},
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 11,
                                       'endline': 11,
                                       'content': '#no space after hash\n',
                                       'value': 'no space after hash'},
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 12,
                                       'endline': 12,
                                       'content': '# comment # with hash inside\n',
                                       'value': 'comment # with hash inside'},
                                      {'instruction': 'RUN',
                                       'startline': 13,
                                       'endline': 13,
                                       'content': 'RUN command1\n',
                                       'value': 'command1'},
                                      {'instruction': 'RUN',
                                       'startline': 14,
                                       'endline': 15,
                                       'content': 'RUN command2 && \\\n    command3\n',
                                       'value': 'command2 &&     command3'},
                                      {'instruction': COMMENT_INSTRUCTION,
                                       'startline': 17,
                                       'endline': 17,
                                       'content': '# interrupt RUN\n',
                                       'value': 'interrupt RUN'},
                                      {'instruction': 'RUN',
                                       'startline': 16,
                                       'endline': 18,
                                       'content': 'RUN command4 && \\\n    command5\n',
                                       'value': 'command4 &&     command5'}]

    def test_invalid_dockerfile_structure(self, dfparser):
        '''Invalid instruction is reserverd.'''
        dfparser.content = dedent("""\
            RUN apt-get update
                apt-get install something
            """)
        assert dfparser.structure == [
                                      {'instruction': 'RUN',
                                       'startline': 0,
                                       'endline': 0,
                                       'content': 'RUN apt-get update\n',
                                       'value': 'apt-get update'},
                                      {'instruction': 'APT-GET',
                                       'startline': 1,
                                       'endline': 1,
                                       'content': '    apt-get install something\n',
                                       'value': 'install something'}]

    def test_dockerfile_json(self, dfparser):
        dfparser.content = dedent("""\
            # comment
            From  base:❤
            LABEL foo="bar❤baz"
            USER  {0}""").format(NON_ASCII)
        expected = json.dumps([{"COMMENT": "comment"},
                               {"FROM": "base:❤"},
                               {"LABEL": "foo=\"bar❤baz\""},
                               {"USER": "{0}".format(NON_ASCII)}])
        assert dfparser.json == expected

    def test_multistage_dockerfile(self, dfparser):
        dfparser.content = dedent("""\
            From builder:image AS builder
            RUN compile to /spam/eggs/❤

            FROM base
            COPY --from=builder /spam/eggs/❤ /usr/bin/eggs
            """)
        expected = [
            {
                'instruction': 'FROM',
                'value': 'builder:image AS builder',
                'startline': 0,  # 0-based
                'endline': 0,
                'content': 'From builder:image AS builder\n',
            }, {
                'instruction': 'RUN',
                'value': 'compile to /spam/eggs/❤',
                'startline': 1,
                'endline': 1,
                'content': 'RUN compile to /spam/eggs/❤\n',
            }, {
                'instruction': 'FROM',
                'value': 'base',
                'startline': 3,
                'endline': 3,
                'content': 'FROM base\n',
            }, {
                'instruction': 'COPY',
                'value': '--from=builder /spam/eggs/❤ /usr/bin/eggs',
                'startline': 4,
                'endline': 4,
                'content': 'COPY --from=builder /spam/eggs/❤ /usr/bin/eggs\n',
            },
        ]
        assert dfparser.structure == expected
        assert dfparser.baseimage == 'base'
        assert dfparser.is_multistage

    def test_multistage_dockerfile_labels(self, dfparser):
        dfparser.content = dedent("""\
            From builder:image AS builder
            LABEL "spam=baked❤beans"

            FROM base
            LABEL "eggs=bacon❤"
            """)
        # only labels from final stage should count
        assert dfparser.labels == {'eggs': 'bacon❤'}

    def test_get_baseimg_from_df(self, dfparser):
        dfparser.lines = ["From fedora:latest\n",
                          "LABEL a b\n"]
        assert dfparser.baseimage == 'fedora:latest'

    def test_get_baseimg_from_df_with_platform(self, dfparser):
        dfparser.lines = ["From --platform=linux/amd64 fedora:latest\n",
                          "LABEL a b\n"]
        assert dfparser.baseimage == 'fedora:latest'

    def test_get_baseimg_from_arg(self, dfparser):
        dfparser.lines = ["ARG BASE=fedora:latest\n",
                          "FROM $BASE\n",
                          "LABEL a b\n"]
        assert dfparser.baseimage == 'fedora:latest'

    def test_get_baseimg_from_build_arg(self, tmpdir):
        tmpdir_path = str(tmpdir.realpath())
        b_args = {"BASE": "fedora:latest"}
        dfp = DockerfileParser(tmpdir_path, env_replace=True, build_args=b_args)
        dfp.lines = ["ARG BASE=centos:latest\n",
                     "FROM $BASE\n",
                     "LABEL a b\n"]
        assert dfp.baseimage == 'fedora:latest'
        assert not dfp.args

    def test_set_no_baseimage(self, dfparser):
        dfparser.lines = []
        with pytest.raises(RuntimeError):
            dfparser.baseimage = 'fedora:latest'
        assert not dfparser.baseimage

    def test_get_build_args(self, tmpdir):
        tmpdir_path = str(tmpdir.realpath())
        b_args = {"bar": "baz❤"}
        df1 = DockerfileParser(tmpdir_path, env_replace=True, build_args=b_args)
        df1.lines = [
            "ARG foo=\"baz❤\"\n",
            "ARG not=\"used\"\n",
            "FROM parent\n",
            "ARG foo\n",
            "ARG bar\n",
            "LABEL label=\"$foo $bar\"\n"
        ]

        # Even though we inherit an ARG, this .args count should only be for the
        # ARGs defined in *this* Dockerfile as we're parsing the Dockerfile and
        # the build_args is only to satisfy use of this build.
        assert len(df1.args) == 2
        assert df1.args.get('foo') == 'baz❤'
        assert df1.args.get('bar') == 'baz❤'
        assert len(df1.labels) == 1
        assert df1.labels.get('label') == 'baz❤ baz❤'

    def test_get_build_args_from_scratch(self, tmpdir):
        tmpdir_path = str(tmpdir.realpath())
        b_args = {"bar": "baz"}
        df1 = DockerfileParser(tmpdir_path, env_replace=True, build_args=b_args)
        df1.lines = [
            "FROM scratch\n",
        ]

        assert not df1.args
        assert not (df1.args == ['bar', 'baz'])
        assert hash(df1.args)

    def test_get_parent_env(self, tmpdir):
        tmpdir_path = str(tmpdir.realpath())
        p_env = {"bar": "baz❤"}
        df1 = DockerfileParser(tmpdir_path, env_replace=True, parent_env=p_env)
        df1.lines = [
            "FROM parent\n",
            "ENV foo=\"$bar\"\n",
            "LABEL label=\"$foo $bar\"\n"
        ]

        # Even though we inherit an ENV, this .envs count should only be for the
        # ENVs defined in *this* Dockerfile as we're parsing the Dockerfile and
        # the parent_env is only to satisfy use of inherited ENVs.
        assert len(df1.envs) == 1
        assert df1.envs.get('foo') == 'baz❤'
        assert len(df1.labels) == 1
        assert df1.labels.get('label') == 'baz❤ baz❤'

    def test_get_parent_env_from_scratch(self, tmpdir):
        tmpdir_path = str(tmpdir.realpath())
        p_env = {"bar": "baz"}
        df1 = DockerfileParser(tmpdir_path, env_replace=True, parent_env=p_env)
        df1.lines = [
            "FROM scratch\n",
        ]

        assert not df1.envs
        assert not (df1.envs == ['bar', 'baz'])
        assert hash(df1.envs)

    @pytest.mark.parametrize(('instr_value', 'expected'), [
        # pylint: disable=anomalous-backslash-in-string
        ('"name1"=\'value 1\' "name2"=myself name3=""',
         {'name1': 'value 1',
          'name2': 'myself',
          'name3': ''}),
        ('name5=5', {'name5': '5'}),
        ('"name6"=6', {'name6': '6'}),
        ('name7', {'name7': ''}),
        ('"name8"', {'name8': ''}),
        ('"name9"="asd \\  \\n qwe"', {'name9': 'asd \\  \\n qwe'}),
        ('"name10"="{0}"'.format(NON_ASCII), {'name10': NON_ASCII}),
        ('"name1 1"=1', {'name1 1': '1'}),
        ('"name12"=12 \\ \n   "name13"=13', {'name12': '12', 'name13': '13'}),
        ('name14=1\\ 4', {'name14': '1 4'}),
        ('name15="with = in value"', {'name15': 'with = in value'}),
        ('name16=❤', {'name16': '❤'}),
        ('name❤=❤', {'name❤': '❤'}),
        # old syntax (without =)
        ('name101 101', {'name101': '101'}),
        ('name102 1 02', {'name102': '1 02'}),
        ('"name103" 1 03', {'name103': '1 03'}),
        ('name104 "1"  04', {'name104': '1  04'}),
        ('name105 1 \'05\'', {'name105': '1 05'}),
        ('name106 1 \'0\'   6', {'name106': '1 0   6'}),
        ('name107 1 0\\ 7', {'name107': '1 0 7'}),
        ('name108 "with = in value"', {'name108': 'with = in value'}),
        ('name109 "\\"quoted\\""', {'name109': '"quoted"'}),
        ('name110 ❤', {'name110': '❤'}),
        ('name1❤ ❤', {'name1❤': '❤'}),
    ])
    def test_get_instructions_from_df(self, dfparser, instruction, instr_value,
                                      expected):
        dfparser.content = "{0} {1}\n".format(instruction, instr_value)
        if instruction == 'LABEL':
            instructions = dfparser.labels
        elif instruction == 'ENV':
            instructions = dfparser.envs
        elif instruction == 'ARG':
            instructions = dfparser.args
        else:
            assert False, 'Unexpected instruction: {0}'.format(instruction)

        assert instructions == expected

    @pytest.mark.parametrize(('from_value', 'expect'), [
        (
            "    ",
            (None, None),
        ), (
            "   foo",
            ('foo', None),
        ), (
            "foo:bar as baz   ",
            ('foo:bar', 'baz'),
        ), (
            "foo as baz",
            ('foo', 'baz'),
        ), (
            "foo and some other junk",  # we won't judge
            ('foo', None),
        ), (
            "registry.example.com:5000/foo/bar:baz",
            ('registry.example.com:5000/foo/bar:baz', None),
        )
    ])
    def test_image_from(self, from_value, expect):
        result = image_from(from_value)
        assert result == expect

    def test_parent_images(self, dfparser):
        FROM = ('my-builder:latest', 'rhel7:7.5')
        template = dedent("""\
            FROM {0} AS builder
            CMD do some stuff

            FROM {1}
            COPY --from=builder some stuff
            """)
        dfparser.content = template.format(*FROM)

        parents = dfparser.parent_images
        assert parents == list(FROM)

        NEW_FROM = ('my-builder@sha256:1234abcd...', 'rhel7@sha256:1234abcd...')
        dfparser.parent_images = NEW_FROM
        assert dfparser.content == template.format(*NEW_FROM)

        with pytest.raises(RuntimeError):
            dfparser.parent_images = [1]
        with pytest.raises(RuntimeError):
            dfparser.parent_images = [1, 2, "many"]

    def test_parent_images_missing_from(self, dfparser):
        dfparser.content = dedent("""\
            # even though this would be really broken
            FROM
            FROM first AS foo
            FROM
            FROM second
            """)
        assert dfparser.parent_images == ['first', 'second']
        assert dfparser.baseimage == 'second'
        dfparser.parent_images = ['spam', 'eggs']
        assert dfparser.parent_images == ['spam', 'eggs']
        # remains just as broken
        assert dfparser.content.count('FROM') == 4

    def test_modify_instruction(self, dfparser):
        FROM = ('ubuntu', 'fedora:❤')
        CMD = ('old❤cmd', 'new❤command')
        df_content = dedent("""\
            FROM {0}
            CMD {1}""").format(FROM[0], CMD[0])

        dfparser.content = df_content

        assert dfparser.baseimage == FROM[0]
        dfparser.baseimage = FROM[1]
        assert dfparser.baseimage == FROM[1]

        assert dfparser.cmd == CMD[0]
        dfparser.cmd = CMD[1]
        assert dfparser.cmd == CMD[1]

    def test_modify_from_multistage(self, dfparser):
        CODE_VERSION = 'latest.❤'
        BASE_FROM = 'base:${CODE_VERSION}'
        BUILDER_FROM = 'builder:${CODE_VERSION}'
        UPDATED_BASE_FROM = 'bass:${CODE_VERSION}'

        BASE_CMD = None
        BUILDER_CMD = '/code/run-extras'
        UPDATED_BASE_CMD = '/code/run-main-actors'

        df_content = dedent("""\
            ARG  CODE_VERSION={0}
            FROM {1}
            CMD {2}

            FROM {3}
            """).format(CODE_VERSION, BUILDER_FROM, BUILDER_CMD, BASE_FROM)

        INDEX_FIRST_FROM = 1
        INDEX_SECOND_FROM = 4

        INDEX_FIRST_CMD = 2
        INDEX_SECOND_CMD = 5

        dfparser.content = df_content

        assert dfparser.baseimage == 'base:{0}'.format(CODE_VERSION)
        assert dfparser.lines[INDEX_FIRST_FROM].strip() == 'FROM {0}'.format(BUILDER_FROM)
        assert dfparser.lines[INDEX_SECOND_FROM].strip() == 'FROM {0}'.format(BASE_FROM)

        dfparser.baseimage = UPDATED_BASE_FROM  # should update only last FROM
        assert dfparser.baseimage == 'bass:{0}'.format(CODE_VERSION)
        assert dfparser.lines[INDEX_FIRST_FROM].strip() == 'FROM {0}'.format(BUILDER_FROM)
        assert dfparser.lines[INDEX_SECOND_FROM].strip() == 'FROM {0}'.format(UPDATED_BASE_FROM)

        assert dfparser.cmd == BASE_CMD  # Last stage command is the base command; None, initially
        assert dfparser.lines[INDEX_FIRST_CMD].strip() == 'CMD {0}'.format(BUILDER_CMD)
        assert len(dfparser.lines) == INDEX_SECOND_CMD

        # Like FROM, updates to CMD should update only the CMD in the final stage.
        dfparser.cmd = UPDATED_BASE_CMD
        assert dfparser.cmd == UPDATED_BASE_CMD
        assert dfparser.lines[INDEX_FIRST_CMD].strip() == 'CMD {0}'.format(BUILDER_CMD)
        assert dfparser.lines[INDEX_SECOND_CMD].strip() == 'CMD {0}'.format(UPDATED_BASE_CMD)

    def test_add_del_instruction(self, dfparser):
        df_content = dedent("""\
            CMD xyz
            LABEL a=b c=d
            LABEL x=\"y z\"
            ENV h i
            ENV j='k' l=m
            ARG a b
            ARG c='d' e=f
            """)
        dfparser.content = df_content

        dfparser._add_instruction('FROM', 'fedora')
        assert dfparser.baseimage == 'fedora'
        dfparser._delete_instructions('FROM')
        assert dfparser.baseimage is None

        dfparser._add_instruction('FROM', 'fedora')
        assert dfparser.baseimage == 'fedora'
        dfparser._delete_instructions('FROM', 'centos')
        assert dfparser.baseimage == 'fedora'
        dfparser._delete_instructions('FROM', 'fedora')
        assert dfparser.baseimage is None

        dfparser._add_instruction('LABEL', ('Name', 'self'))
        assert len(dfparser.labels) == 4
        assert dfparser.labels.get('Name') == 'self'
        dfparser._delete_instructions('LABEL')
        assert dfparser.labels == {}

        dfparser._add_instruction('ENV', ('Name', 'self'))
        assert len(dfparser.envs) == 4
        assert dfparser.envs.get('Name') == 'self'
        dfparser._delete_instructions('ENV')
        assert dfparser.envs == {}

        dfparser._add_instruction('ARG', ('Name', 'self'))
        assert len(dfparser.args) == 4
        assert dfparser.args.get('Name') == 'self'
        dfparser._delete_instructions('ARG')
        assert dfparser.envs == {}

        assert dfparser.cmd == 'xyz'

    @pytest.mark.parametrize(('existing',
                              'delete_key',
                              'expected'), [
        # Delete non-existing key
        (['a b\n',
          'x="y z"\n'],
         'name',
         KeyError()),

        # Simple remove
        (['a b\n',
          'x="y z"\n'],
         'a',
         ['x="y z"\n']),

        # Simple remove
        (['a b\n',
          'x="y z"\n'],
         'x',
         ['a b\n']),

        # Simple remove unicode
        (['a b\n',
          'x="y ❤"\n'],
         'x',
         ['a b\n']),

        # Simple remove unicode
        (['a b\n',
          '❤="y z"\n'],
         '❤',
         ['a b\n']),

        #  Remove first of two instructions on the same line
        (['a b\n',
          'x="y z"\n',
          '"first"="first" "second"="second"\n'],
         'first',
         ['a b\n',
          'x="y z"\n',
          '"second"="second"\n']),

        #  Remove second of two instructions on the same line
        (['a b\n',
          'x="y z"\n',
          '"first"="first" "second"="second"\n'],
         'second',
         ['a b\n',
          'x="y z"\n',
          '"first"="first"\n']),
    ])
    def test_delete_instruction(self, dfparser, instruction, existing, delete_key, expected):
        existing = [instruction + ' ' + i for i in existing]
        if isinstance(expected, list):
            expected = [instruction + ' ' + i for i in expected]
        dfparser.lines = ["FROM xyz\n"] + existing

        if isinstance(expected, KeyError):
            with pytest.raises(KeyError):
                dfparser._delete_instructions(instruction, delete_key)
        else:
            dfparser._delete_instructions(instruction, delete_key)
            assert set(dfparser.lines[1:]) == set(expected)

    @pytest.mark.parametrize(('existing',
                              'new',
                              'expected'), [
        # Simple test: set an instruction
        (['a b\n',
          'x="y z"\n'],
         {'Name': 'New shiny project'},
         ['Name=\'New shiny project\'\n']),

        # Set two instructions
        (['a b\n',
          'x="y z"\n'],
         {'something': 'nothing', 'mine': 'yours'},
         ['something=nothing\n', 'mine=yours\n']),

        # Set instructions to what they already were: should be no difference
        (['a b\n',
          'x="y z"\n',
          '"first"="first" second=\'second value\'\n'],
         {'a': 'b', 'x': 'y z', 'first': 'first', 'second': 'second value'},
         ['a b\n',
          'x="y z"\n',
          '"first"="first" second=\'second value\'\n']),

        # Adjust one label of a multi-value LABEL/ENV statement
        (['a b\n',
          'first=\'first value\' "second"=second\n',
          'x="y z"\n'],
         {'first': 'changed', 'second': 'second'},
         ['first=changed "second"=second\n']),

        # Delete one label of a multi-value LABEL/ENV statement
        (['a b\n',
          'x="y z"\n',
          'first=first second=second\n'],
         {'second': 'second'},
         ['second=second\n']),

        # Nested quotes
        (['"ownership"="Alice\'s label" other=value\n'],
         {'ownership': "Alice's label"},
         # Keeps existing key quoting style
         ['"ownership"="Alice\'s label"\n']),

        # Modify a single value that needs quoting
        (['foo bar\n'],
         {'foo': 'extra bar'},
         ["foo 'extra bar'\n"]),
    ])
    def test_setter(self, dfparser, instruction, existing, new, expected):
        existing = [instruction + ' ' + i for i in existing]
        if isinstance(expected, list):
            expected = [instruction + ' ' + i for i in expected]
        dfparser.lines = ["FROM xyz\n"] + existing

        if instruction == 'LABEL':
            dfparser.labels = new
            assert dfparser.labels == new
        elif instruction == 'ENV':
            dfparser.envs = new
            assert dfparser.envs == new
        elif instruction == 'ARG':
            dfparser.args = new
            assert dfparser.args == new
        assert set(dfparser.lines[1:]) == set(expected)

    @pytest.mark.parametrize(('old_instructions', 'key', 'new_value', 'expected'), [
        # Simple case, no '=' or quotes
        ('Release 1', 'Release', '2', 'Release 2'),
        # No '=' but quotes (which are kept)
        ('"Release" "2"', 'Release', '3', '"Release" 3'),
        # Simple case, '=' but no quotes
        ('Release=1', 'Release', '6', 'Release=6'),
        # '=' and quotes, with space in the value
        ('"Name"=\'alpha alpha\' Version=1',
         'Name', 'beta delta', '"Name"=\'beta delta\' Version=1'),
        ('Name=foo', 'Name', 'new value', "Name='new value'"),
        # ' ' and quotes
        ('"Name" alpha alpha', 'Name', 'beta delta', "\"Name\" 'beta delta'"),
        # '=', multiple labels, no quotes
        ('Name=foo Release=3', 'Release', '4', 'Name=foo Release=4'),
        # '=', multiple labels and quotes
        ('Name=\'foo bar\' "Release"="4"', 'Release', '5', 'Name=\'foo bar\' "Release"=5'),
        # Release that's not entirely numeric
        ('Version=1.1', 'Version', '2.1', 'Version=2.1'),
    ])
    def test_setter_direct(self, dfparser, instruction, old_instructions, key, new_value, expected):
        df_content = dedent("""\
            FROM xyz
            LABEL a b
            LABEL x=\"y z\"
            ENV c d
            ENV e=\"f g\"
            {0} {1}
            """).format(instruction, old_instructions)

        dfparser.content = df_content
        if instruction == 'LABEL':
            dfparser.labels[key] = new_value
            assert dfparser.labels[key] == new_value
            assert dfparser.lines[-1] == '{0} {1}\n'.format(instruction, expected)
            del dfparser.labels[key]
            assert not dfparser.labels.get(key)
        elif instruction == 'ENV':
            dfparser.envs[key] = new_value
            assert dfparser.envs[key] == new_value
            assert dfparser.lines[-1] == '{0} {1}\n'.format(instruction, expected)
            del dfparser.envs[key]
            assert not dfparser.labels.get(key)

    @pytest.mark.parametrize('instruction', ('ARG', 'ENV'))
    @pytest.mark.parametrize('separator', [' ', '='])
    @pytest.mark.parametrize(('label', 'expected'), [
        # Expected substitutions
        ('$V', 'v'),
        ('"$V"', 'v'),
        ('$V-foo', 'v-foo'),
        ('"$V-foo"', 'v-foo'),
        ('"$V"-foo', 'v-foo'),
        ('${V}', 'v'),
        ('${V}-foo', 'v-foo'),
        ('$V-{foo}', 'v-{foo}'),
        ('$V-❤', 'v-❤'),
        ('$VS', 'spam maps'),

        # These should not be substituted, only dequoted
        ("'$V'", "$V"),
        ("\\$V", "$V"),
        ("\\$V❤", "$V❤"),

        # Try to trip up the parser
        ('\\"$V', '"v'),
        ("\\'$V", "'v"),
        ('$V}', 'v}'),
        ('${}', ''),
        ("'\\'$V'\\'", "\\v\\"),
    ])
    def test_arg_env_replace(self, dfparser, instruction, separator, label, expected):
        dfparser.lines = ["FROM fedora\n",
                          "{0} V=v\n".format(instruction),
                          "{0} VS='spam maps'\n".format(instruction),
                          "LABEL TEST{0}{1}\n".format(separator, label)]
        assert dfparser.labels['TEST'] == expected
        with pytest.raises(TypeError):
            dfparser.labels = ['foo', 'bar']

    @pytest.mark.parametrize('instruction', ('ARG', 'ENV'))
    @pytest.mark.parametrize('separator', [' ', '='])
    @pytest.mark.parametrize(('label', 'expected'), [
        # These would have been substituted with env_replace=True
        ('$V', '$V'),
        ('"$V"', '$V'),
        ('$V-foo', '$V-foo'),
        ('"$V-foo"', '$V-foo'),
        ('"$V"-foo', '$V-foo'),
        ('"$V"-❤', '$V-❤'),
    ])
    def test_arg_env_noreplace(self, dfparser, instruction, separator, label, expected):
        """
        Make sure environment replacement can be disabled.
        """
        dfparser.env_replace = False
        dfparser.lines = ["FROM fedora\n",
                          "{0} V=v\n".format(instruction),
                          "LABEL TEST{0}{1}\n".format(separator, label)]
        assert dfparser.labels['TEST'] == expected

    @pytest.mark.parametrize('instruction', ('ARG', 'ENV'))
    @pytest.mark.parametrize('label', [
        '${V',
        '"${V"',
        '${{{{V}',
    ])
    def test_arg_env_invalid(self, dfparser, instruction, label):
        """
        These tests are invalid, but the parser should at least terminate
        even if it raises an exception.
        """
        dfparser.lines = ["FROM fedora\n",
                          "{0} v=v\n".format(instruction),
                          "LABEL TEST={0}\n".format(label)]
        try:
            dfparser.labels['TEST']
        except KeyError:
            pass

    @pytest.mark.parametrize(('instruction', 'attribute'), (
        ('ARG', 'args'),
        ('ENV', 'envs'),
    ))
    def test_arg_env_multistage(self, dfparser, instruction, attribute):
        dfparser.content = dedent("""\
            FROM stuff
            {instruction} a=keep❤ b=keep❤

            FROM base
            {instruction} a=delete❤
            RUN something
        """.format(instruction=instruction))

        getattr(dfparser, attribute)['a'] = "changed❤"
        del getattr(dfparser, attribute)['a']
        getattr(dfparser, attribute)['b'] = "new❤"

        lines = dfparser.lines
        assert instruction in lines[1]
        assert "a=keep❤" in lines[1]
        assert "b=new❤" not in lines[1]
        assert "a=delete❤" not in dfparser.content
        assert "b='new❤'" in lines[-1]  # unicode quoted

    @pytest.mark.xfail
    @pytest.mark.parametrize('instruction', ('ARG', 'ENV'))
    @pytest.mark.parametrize(('label', 'expected'), [
        ('${V:-foo}', 'foo'),
        ('${V:+foo}', 'v'),
        ('${UNDEF:+foo}', 'foo'),
        ('${UNDEF:+${V}}', 'v'),
    ])
    def test_arg_env_replace_notimplemented(self, dfparser, instruction, label, expected):
        """
        Test for syntax we don't support yet but should.
        """
        dfparser.lines = ["FROM fedora\n",
                          "{0} V=v\n".format(instruction),
                          "LABEL TEST={0}\n".format(label)]
        assert dfparser.labels['TEST'] == expected

    def test_path_and_fileobj_together(self):
        with pytest.raises(ValueError):
            DockerfileParser(path='.', fileobj=io.StringIO())

    def test_nonseekable_fileobj(self):
        with pytest.raises((AttributeError, io.UnsupportedOperation)):
            DockerfileParser(fileobj=sys.stdin)

    def test_context_structure_per_line(self, dfparser, instruction):
        dfparser.content = dedent("""\
            FROM fedora:25

            {0} multi.label❤1="value❤1" \\
                  multi.label❤2="value❤2" \\
                  other="value❤3"

            {0} 2multi.label1="othervalue1" 2multi.label2="othervalue2" other="othervalue3"

            {0} "com.example.vendor"="ACME Incorporated"
            {0} com.example.label-with-value="foo"
            {0} version="1.0.❤"
            {0} description="This text illustrates ❤ \\
            that label-values can span multiple lines."
            {0} key="with = in the value❤"
            """).format(instruction)

        c = dfparser.context_structure

        assert c[1].get_line_value(context_type=instruction) == {
            "multi.label❤1": "value❤1",
            "multi.label❤2": "value❤2",
            "other": "value❤3"
        }

        assert c[2].get_line_value(context_type=instruction) == {
            "2multi.label1": "othervalue1",
            "2multi.label2": "othervalue2",
            "other": "othervalue3"
        }

        assert c[3].get_line_value(context_type=instruction) == {
            "com.example.vendor": "ACME Incorporated"
        }

        assert c[4].get_line_value(context_type=instruction) == {
            "com.example.label-with-value": "foo"
        }

        assert c[5].get_line_value(context_type=instruction) == {
            "version": "1.0.❤"
        }

        assert c[6].get_line_value(context_type=instruction) == {
            "description": "This text illustrates ❤ that label-values can span multiple lines."
        }

        assert c[7].get_line_value(context_type=instruction) == {
            "key": "with = in the value❤"
        }

    def test_context_structure(self, dfparser, instruction):
        dfparser.content = dedent("""\
            FROM fedora:25

            {0} multi.label❤1="value❤1" \\
                  multi.label❤2="value❤2" \\
                  other="value❤3"

            {0} 2multi.label1="othervalue1" 2multi.label2="othervalue2" other="othervalue3"

            {0} "com.example.vendor"="ACME Incorporated"
            {0} com.example.label-with-value="foo"
            {0} version="1.0.❤"
            {0} description="This text illustrates \\
            that label-values can span multiple lines."
            """).format(instruction)

        c = dfparser.context_structure

        assert c[1].get_values(context_type=instruction) == {
            "multi.label❤1": "value❤1",
            "multi.label❤2": "value❤2",
            "other": "value❤3"
        }

        assert c[2].get_values(context_type=instruction) == {
            "multi.label❤1": "value❤1",
            "multi.label❤2": "value❤2",
            "other": "othervalue3",
            "2multi.label1": "othervalue1",
            "2multi.label2": "othervalue2"
        }

        assert c[3].get_values(context_type=instruction) == {
            "multi.label❤1": "value❤1",
            "multi.label❤2": "value❤2",
            "other": "othervalue3",
            "2multi.label1": "othervalue1",
            "2multi.label2": "othervalue2",
            "com.example.vendor": "ACME Incorporated"
        }

        assert c[4].get_values(context_type=instruction) == {
            "multi.label❤1": "value❤1",
            "multi.label❤2": "value❤2",
            "other": "othervalue3",
            "2multi.label1": "othervalue1",
            "2multi.label2": "othervalue2",
            "com.example.vendor": "ACME Incorporated",
            "com.example.label-with-value": "foo"
        }

        assert c[5].get_values(context_type=instruction) == {
            "multi.label❤1": "value❤1",
            "multi.label❤2": "value❤2",
            "other": "othervalue3",
            "2multi.label1": "othervalue1",
            "2multi.label2": "othervalue2",
            "com.example.vendor": "ACME Incorporated",
            "com.example.label-with-value": "foo",
            "version": "1.0.❤"
        }

        assert c[6].get_values(context_type=instruction) == {
            "multi.label❤1": "value❤1",
            "multi.label❤2": "value❤2",
            "other": "othervalue3",
            "2multi.label1": "othervalue1",
            "2multi.label2": "othervalue2",
            "com.example.vendor": "ACME Incorporated",
            "com.example.label-with-value": "foo",
            "version": "1.0.❤",
            "description": "This text illustrates that label-values can span multiple lines."
        }

    def test_context_structure_mixed(self, dfparser, instruction):
        dfparser.content = dedent("""\
            FROM fedora:25

            {0} key=value❤
            RUN touch /tmp/a
            {0} key2=value2❤""").format(instruction)

        c = dfparser.context_structure
        assert c[0].get_values(context_type=instruction) == {}
        assert c[1].get_values(context_type=instruction) == {"key": "value❤"}
        assert c[2].get_values(context_type=instruction) == {"key": "value❤"}
        assert c[3].get_values(context_type=instruction) == {"key": "value❤",
                                                             "key2": "value2❤"}

    @pytest.mark.parametrize('instruction', ('ARG', 'ENV'))
    def test_context_structure_mixed_arg_env_label(self, dfparser, instruction):
        dfparser.content = dedent("""\
            FROM fedora:25

            {0} key=value❤
            RUN touch /tmp/a
            LABEL key2=value2❤""".format(instruction))
        c = dfparser.context_structure

        assert c[0].get_values(context_type=instruction) == {}
        assert c[0].get_values(context_type="LABEL") == {}

        assert c[1].get_values(context_type=instruction) == {"key": "value❤"}
        assert c[1].get_values(context_type="LABEL") == {}

        assert c[2].get_values(context_type=instruction) == {"key": "value❤"}
        assert c[2].get_values(context_type="LABEL") == {}

        assert c[3].get_values(context_type=instruction) == {"key": "value❤"}
        assert c[3].get_values(context_type="LABEL") == {"key2": "value2❤"}

    def test_context_structure_mixed_top_arg(self, tmpdir):
        dfp = DockerfileParser(
            str(tmpdir.realpath()),
            build_args={"version": "8", "key": "value❤"},
            env_replace=True)
        dfp.content = dedent("""\
            ARG image=centos
            ARG version=latest
            FROM $image:$version
            ARG image
            ARG key
            """)
        c = dfp.context_structure

        assert len(c) == 5
        assert c[0].get_values(context_type='ARG') == {"image": "centos"}
        assert c[1].get_values(context_type='ARG') == {"image": "centos", "version": "8"}
        assert c[2].get_values(context_type='ARG') == {}
        assert c[3].get_values(context_type='ARG') == {"image": "centos"}
        assert c[4].get_values(context_type='ARG') == {"image": "centos", "key": "value❤"}

    def test_expand_concatenated_variables(self, dfparser):
        dfparser.content = dedent("""\
            FROM scratch
            ENV NAME=name VER=1
            LABEL component="$NAME$VER❤"
        """)
        assert dfparser.labels['component'] == 'name1❤'

    @pytest.mark.parametrize('instruction', ('ARG', 'ENV'))
    def test_label_arg_env_key(self, dfparser, instruction):
        """
        Verify keys may be substituted with values containing space.

        Surprisingly, Docker allows environment variable substitution even
        in the keys of labels, and not only that but it allows them to
        contain spaces.
        """
        dfparser.content = dedent("""\
            FROM scratch
            {0} FOOBAR="foo bar"
            LABEL "$FOOBAR"="baz"
        """.format(instruction))
        assert dfparser.labels['foo bar'] == 'baz'

    @pytest.mark.parametrize('label_value, bad_keyval, envs', [
        ('a=b c', 'c', None),
        ('a=b ❤', '❤', None),
        # if variable substitution was done too early, this could be an issue
        ('a=1 $CHEEKY_VARIABLE', '$CHEEKY_VARIABLE', {'CHEEKY_VARIABLE': 'b=2'})
    ])
    @pytest.mark.parametrize('action', ['get', 'set'])
    def test_label_invalid(self, dfparser, label_value, bad_keyval, envs, action):
        if envs:
            env_vals = ('{0}="{1}"'.format(k, v) for k, v in envs.items())
            env_line = 'ENV {values}\n'.format(values=' '.join(env_vals))
        else:
            env_line = ''

        dfparser.lines = [
            "FROM scratch\n",
            env_line,   # has to appear before the LABEL line
            "LABEL {0}\n".format(label_value),
        ]
        with pytest.raises(ValueError) as exc_info:
            if action == 'get':
                dfparser.labels  # pylint: disable=pointless-statement
            elif action == 'set':
                dfparser.labels = {}
        msg = exc_info.value.args[0]
        assert msg == ('Syntax error - can\'t find = in "{word}". '
                       'Must be of the form: name=value'
                       .format(word=bad_keyval))

    def test_add_lines_stages(self, dfparser):
        dfparser.content = dedent("""\
            From builder
            CMD xyz ❤
            From base
            LABEL a=b c=d
            ENV h i
            """)
        dfparser.add_lines("something new ❤", all_stages=True)
        assert "something new ❤" in dfparser.lines[2]
        assert "something new ❤" in dfparser.lines[-1]
        assert len([line for line in dfparser.lines if "something new ❤" in line]) == 2
        assert len(dfparser.lines) == 7

    @pytest.mark.parametrize('at_start', [True, False])
    def test_add_lines_stages_skip_scratch(self, dfparser, at_start):
        dfparser.content = dedent("""\
            From builder
            CMD xyz ❤
            From scratch
            LABEL type=scratch
            From base
            LABEL a=b c=d
            ENV h i
            From scratch
            LABEL type=scratch2
            From scratch as foo
            LABEL type=scratch3
            """)
        dfparser.add_lines("something new ❤", all_stages=True, skip_scratch=True, at_start=at_start)

        if at_start:
            assert "something new ❤" in dfparser.lines[1]
            assert "something new ❤" in dfparser.lines[6]
        else:
            assert "something new ❤" in dfparser.lines[2]
            assert "something new ❤" in dfparser.lines[8]
        assert len([line for line in dfparser.lines if "something new ❤" in line]) == 2
        assert len(dfparser.lines) == 13

    def test_add_lines_stage_edge(self, dfparser):
        dfparser.content = "# no from or newline ❤"
        dfparser.add_lines("begin with new ❤", at_start=True)
        dfparser.add_lines("end with new ❤")
        assert "begin with new ❤" in dfparser.lines[0]
        assert "end with new ❤" in dfparser.lines[2]

    @pytest.mark.parametrize(('anchor', 'raises'), [
        (
            3, None
        ),
        (
            'CMD xyz ❤\n', None
        ),
        (
            dict(
                content='CMD xyz ❤\n',
                startline=3,
                endline=3,
                instruction='CMD',
                value='xyz ❤'
            ),
            None
        ),
        (
            -2, AssertionError
        ),
        (
            20, AssertionError
        ),
        (
            2.0, RuntimeError
        ),
        (
            'not there', RuntimeError
        ),
        (
            dict(), AssertionError
        ),
    ])
    def test_add_lines_at(self, dfparser, anchor, raises):
        dfparser.content = dedent("""\
            From builder
            CMD xyz ❤
            LABEL a=b c=d
            CMD xyz ❤
            """)

        if raises:
            with pytest.raises(raises):
                dfparser.add_lines_at(anchor, "# something new ❤")
            return

        dfparser.add_lines_at(anchor, "# something new ❤")
        assert "something new ❤" in dfparser.content
        assert "something new ❤" not in dfparser.lines[1]
        assert "something new ❤" in dfparser.lines[3]
        assert "CMD" in dfparser.lines[4]

    @pytest.mark.parametrize('anchor', [
        1,
        'CMD xyz ❤\n',
        dict(
            content='CMD xyz ❤\n',
            startline=1,
            endline=1,
            instruction='CMD',
            value='xyz ❤'
        ),
    ])
    def test_replace_lines_at(self, dfparser, anchor):
        dfparser.content = dedent("""\
            From builder
            CMD xyz ❤
            LABEL a=b c=d
            """)

        dfparser.add_lines_at(anchor, "# something new ❤", replace=True)
        assert "something new ❤" in dfparser.lines[1]
        assert "CMD" not in dfparser.content

    @pytest.mark.parametrize('anchor', [
        1,
        'CMD xyz ❤\n',
        dict(
            content='CMD xyz ❤\n',
            startline=1,
            endline=1,
            instruction='CMD',
            value='xyz ❤'
        ),
    ])
    def test_add_lines_after(self, dfparser, anchor):
        dfparser.content = dedent("""\
            From builder
            CMD xyz ❤
            LABEL a=b c=d
            """)

        dfparser.add_lines_at(anchor, "# something new ❤", after=True)
        assert "something new ❤" in dfparser.lines[2]
        assert "CMD" in dfparser.lines[1]
        assert "LABEL" in dfparser.lines[3]

    def test_add_lines_at_edge(self, dfparser):
        dfparser.content = dedent("""\
            From builder
            CMD xyz ❤
            LABEL a=b c=d e=❤""")  # no newline
        dfparser.add_lines_at(2, "# something new ❤", after=True)
        assert "d#" not in dfparser.content
        assert len(dfparser.lines) == 4
        assert "something new ❤" in dfparser.lines[3]

    def test_add_lines_after_continuation(self, dfparser):
        dfparser.content = dedent("""\
            FROM builder
            RUN touch foo ❤; \\
                touch bar
            """)

        fromline = dfparser.structure[1]
        assert fromline['instruction'] == 'RUN'
        dfparser.add_lines_at(fromline, "# something new", after=True)
        assert dfparser.lines == [
            "FROM builder\n",
            "RUN touch foo ❤; \\\n",
            "    touch bar\n",
            "# something new\n",
        ]

    def test_replace_lines_continuation(self, dfparser):
        dfparser.content = dedent("""\
            FROM builder
            RUN touch foo; \\
                touch bar ❤
            """)

        fromline = dfparser.structure[1]
        assert fromline['instruction'] == 'RUN'
        dfparser.add_lines_at(fromline, "# ❤ something new", replace=True)
        assert dfparser.lines == [
            "FROM builder\n",
            "# ❤ something new\n",
        ]

    def test_remove_whitespace(self, tmpdir):
        """
        Verify keys are parsed correctly even if there is no final newline.

        """
        with open(os.path.join(str(tmpdir), 'Dockerfile'), 'w') as fp:
            fp.write('FROM scratch')
        tmpdir_path = str(tmpdir.realpath())
        df1 = DockerfileParser(tmpdir_path)
        df1.labels['foo'] = 'bar ❤'

        df2 = DockerfileParser(tmpdir_path, True)
        assert df2.baseimage == 'scratch'
        assert df2.labels['foo'] == 'bar ❤'

    def _test_escape_directive(self, dfparser, escape_value, used_line_continuation):
        dfparser.content = dedent("""\
            #    escape=   {escape_value}
            FROM base
            RUN touch foo; {line_cont}
                touch bar
            """.format(escape_value=escape_value, line_cont=used_line_continuation))
        assert dfparser.structure == [
            {
                'instruction': COMMENT_INSTRUCTION,
                'startline': 0,
                'endline': 0,
                'content': '#    escape=   {escape_value}\n'.format(
                    escape_value=escape_value
                ),
                'value': 'escape=   {escape_value}'.format(
                    escape_value=escape_value
                )
            },
            {
                'instruction': 'FROM',
                'startline': 1,
                'endline': 1,
                'content': 'FROM base\n',
                'value': 'base'
            },
            {
                'instruction': 'RUN',
                'startline': 2,
                'endline': 3,
                'content': 'RUN touch foo; {line_cont}\n    touch bar\n'.format(
                                    line_cont=used_line_continuation
                ),
                'value': 'touch foo;     touch bar'
            }
        ]

    @pytest.mark.parametrize(('escape_value', 'used_line_continuation'), [
        ('\\', '\\'),
        ('`', '`'),
    ])
    def test_escape_directive(self, dfparser, escape_value, used_line_continuation):
        self._test_escape_directive(dfparser, escape_value, used_line_continuation)

    @pytest.mark.xfail
    @pytest.mark.parametrize(('escape_value', 'used_line_continuation'), [
        ('\\', '`'),
        ('`', '\\')
    ])
    def test_escape_directive_xfail(self, dfparser, escape_value, used_line_continuation):
        self._test_escape_directive(dfparser, escape_value, used_line_continuation)

    def _test_escape_after_syntax_directive(self, dfparser, escape_value, used_line_continuation):
        dfparser.content = dedent("""\
            # syntax=ubuntu
            #    escape=   {escape_value}
            FROM base
            RUN touch foo; {line_cont}
                touch bar
            """.format(escape_value=escape_value, line_cont=used_line_continuation))
        assert dfparser.structure == [
            {
                'instruction': COMMENT_INSTRUCTION,
                'startline': 0,
                'endline': 0,
                'content': '# syntax=ubuntu\n',
                'value': 'syntax=ubuntu',
            },
            {
                'instruction': COMMENT_INSTRUCTION,
                'startline': 1,
                'endline': 1,
                'content': '#    escape=   {escape_value}\n'.format(
                    escape_value=escape_value
                ),
                'value': 'escape=   {escape_value}'.format(
                    escape_value=escape_value
                )
            },
            {
                'instruction': 'FROM',
                'startline': 2,
                'endline': 2,
                'content': 'FROM base\n',
                'value': 'base'
            },
            {
                'instruction': 'RUN',
                'startline': 3,
                'endline': 4,
                'content': 'RUN touch foo; {line_cont}\n    touch bar\n'.format(
                                    line_cont=used_line_continuation
                ),
                'value': 'touch foo;     touch bar'
            }
        ]

    @pytest.mark.parametrize(('escape_value', 'used_line_continuation'), [
        ('\\', '\\'),
        ('`', '`'),
    ])
    def test_escape_after_syntax_directive(self, dfparser, escape_value, used_line_continuation):
        self._test_escape_after_syntax_directive(dfparser, escape_value, used_line_continuation)

    @pytest.mark.xfail
    @pytest.mark.parametrize(('escape_value', 'used_line_continuation'), [
        ('\\', '`'),
        ('`', '\\'),
    ])
    def test_escape_after_syntax_directive_xfail(
            self,
            dfparser,
            escape_value,
            used_line_continuation
    ):
        self._test_escape_after_syntax_directive(dfparser, escape_value, used_line_continuation)

    def test_escape_directive_ignore_after_comment(self, dfparser):
        dfparser.content = dedent("""\
            # comment
            # escape=`
            FROM base
            RUN touch foo; \\
                touch bar
            """)
        assert dfparser.structure == [
            {
                'instruction': COMMENT_INSTRUCTION,
                'startline': 0,
                'endline': 0,
                'content': '# comment\n',
                'value': 'comment',
            },
            {
                'instruction': COMMENT_INSTRUCTION,
                'startline': 1,
                'endline': 1,
                'content': '# escape=`\n',
                'value': 'escape=`',
            },
            {
                'instruction': 'FROM',
                'startline': 2,
                'endline': 2,
                'content': 'FROM base\n',
                'value': 'base'
            },
            {
                'instruction': 'RUN',
                'startline': 3,
                'endline': 4,
                'content': 'RUN touch foo; \\\n    touch bar\n',
                'value': 'touch foo;     touch bar'
            }
        ]



#!/usr/bin/python
"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

import re
import io

from os import path
from setuptools import setup, find_packages


def _get_requirements(path):
    try:
        with open(path) as f:
            packages = f.read().splitlines()
    except (IOError, OSError) as ex:
        raise RuntimeError("Can't open file with requirements: %s", repr(ex))
    return [p.strip() for p in packages if not re.match(r"^\s*#", p)]


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with io.open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='dockerfile-parse',
    version='2.0.0',
    description='Python library for Dockerfile manipulation',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jiri Popelka',
    author_email='jpopelka@redhat.com',
    url='https://github.com/containerbuildsystem/dockerfile-parse',
    license="BSD",
    packages=find_packages(exclude=["tests"]),
    python_requires='>=3.6',
    install_requires=[],
    tests_require=_get_requirements('tests/requirements.txt'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



# -*- coding: utf-8 -*-
"""
Copyright (c) 2015, 2018, 2019 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

import json
import logging
import os
import re
from contextlib import contextmanager
from shlex import quote

from .constants import DOCKERFILE_FILENAME, COMMENT_INSTRUCTION
from .util import (b2u, extract_key_values, get_key_val_dictionary,
                   u2b, Context, WordSplitter)


logger = logging.getLogger(__name__)


class KeyValues(dict):
    """
    Abstract base class for allowing direct write access to Dockerfile
    instructions which result in a set of key value pairs.

    Subclasses must override the `parser_attr` value.
    """
    parser_attr = None

    def __init__(self, key_values, parser):
        super(KeyValues, self).__init__(key_values)
        self.parser = parser

    def __delitem__(self, key):
        super(KeyValues, self).__delitem__(key)
        setattr(self.parser, self.parser_attr, dict(self))

    def __setitem__(self, key, value):
        super(KeyValues, self).__setitem__(key, value)
        setattr(self.parser, self.parser_attr, dict(self))

    def __eq__(self, other):
        if not isinstance(other, dict):
            return False
        return dict(self) == other

    def __hash__(self):
        return hash(json.dumps(self, separators=(',', ':'), sort_keys=True))


class Labels(KeyValues):
    """
    A class for allowing direct write access to Dockerfile labels, e.g.:

    parser.labels['label'] = 'value'
    """
    parser_attr = 'labels'


class Envs(KeyValues):
    """
    A class for allowing direct write access to Dockerfile env. vars., e.g.:

    parser.envs['variable_name'] = 'value'
    """
    parser_attr = 'envs'


class Args(KeyValues):
    """
    A class for allowing direct write access to Dockerfile build args, e.g.:

    parser.args['variable_name'] = 'value'
    """
    parser_attr = 'args'


class DockerfileParser(object):
    def __init__(self, path=None,
                 cache_content=False,
                 env_replace=True,
                 parent_env=None,
                 fileobj=None,
                 build_args=None):
        """
        Initialize source of Dockerfile
        :param path: path to (directory with) Dockerfile; if not provided,
                     and fileobj is not provided, the current working
                     directory will be used
        :param cache_content: cache Dockerfile content inside DockerfileParser
        :param env_replace: return content with variables replaced
        :param parent_env: python dict of inherited env vars from parent image
        :param fileobj: seekable file-like object containing Dockerfile content
                        as bytes (will be truncated on write)
        :param build_args: python dict of build args used when building image
        """

        self.fileobj = fileobj

        if self.fileobj is not None:
            if path is not None:
                raise ValueError("Parameters path and fileobj cannot be used together.")
            else:
                self.fileobj.seek(0)
        else:
            path = path or '.'
            if path.endswith(DOCKERFILE_FILENAME):
                self.dockerfile_path = path
            else:
                self.dockerfile_path = os.path.join(path, DOCKERFILE_FILENAME)

        self.cache_content = cache_content
        self.cached_content = ''  # unicode string

        if cache_content:
            try:
                # this will cache the Dockerfile content
                self.content
            except (IOError, OSError):
                # the Dockerfile doesn't exist yet
                pass

        self.env_replace = env_replace

        if parent_env is None:
            self.parent_env = {}
        else:
            assert isinstance(parent_env, dict)
            logger.debug("Setting inherited parent image ENV vars: %s", parent_env)
            self.parent_env = parent_env

        if build_args is None:
            self.build_args = {}
        else:
            assert isinstance(build_args, dict)
            logger.debug("Setting build args: %s", build_args)
            self.build_args = build_args

    @contextmanager
    def _open_dockerfile(self, mode):
        if self.fileobj is not None:
            self.fileobj.seek(0)
            if 'w' in mode:
                self.fileobj.truncate()
            yield self.fileobj
            self.fileobj.seek(0)
        else:
            with open(self.dockerfile_path, mode) as dockerfile:
                yield dockerfile

    @property
    def lines(self):
        """
        :return: list containing lines (unicode) from Dockerfile
        """
        if self.cache_content and self.cached_content:
            return self.cached_content.splitlines(True)

        try:
            with self._open_dockerfile('rb') as dockerfile:
                lines = [b2u(line) for line in dockerfile.readlines()]
                if self.cache_content:
                    self.cached_content = ''.join(lines)
                return lines
        except (IOError, OSError) as ex:
            logger.error("Couldn't retrieve lines from dockerfile: %r", ex)
            raise

    @lines.setter
    def lines(self, lines):
        """
        Fill Dockerfile content with specified lines
        :param lines: list of lines to be written to Dockerfile
        """
        if self.cache_content:
            self.cached_content = ''.join(b2u(line) for line in lines)

        try:
            with self._open_dockerfile('wb') as dockerfile:
                dockerfile.writelines(u2b(line) for line in lines)
        except (IOError, OSError) as ex:
            logger.error("Couldn't write lines to dockerfile: %r", ex)
            raise

    @property
    def content(self):
        """
        :return: string (unicode) with Dockerfile content
        """
        if self.cache_content and self.cached_content:
            return self.cached_content

        try:
            with self._open_dockerfile('rb') as dockerfile:
                content = b2u(dockerfile.read())
                if self.cache_content:
                    self.cached_content = content
                return content
        except (IOError, OSError) as ex:
            logger.error("Couldn't retrieve content of dockerfile: %r", ex)
            raise

    @content.setter
    def content(self, content):
        """
        Overwrite Dockerfile with specified content
        :param content: string to be written to Dockerfile
        """
        if self.cache_content:
            self.cached_content = b2u(content)

        try:
            with self._open_dockerfile('wb') as dockerfile:
                dockerfile.write(u2b(content))
        except (IOError, OSError) as ex:
            logger.error("Couldn't write content to dockerfile: %r", ex)
            raise

    @property
    def structure(self):
        """
        Returns a list of dicts describing the commands:
        [
            {"instruction": "FROM",       # always upper-case
             "startline": 0,              # 0-based
             "endline": 0,                # 0-based
             "content": "From fedora\n",
             "value": "fedora"},

            {"instruction": "CMD",
             "startline": 1,
             "endline": 2,
             "content": "CMD yum -y update && \\\n    yum clean all\n",
             "value": "yum -y update && yum clean all"}
        ]
        """
        def _rstrip_eol(text, line_continuation_char='\\'):
            text = text.rstrip()
            if text.endswith(line_continuation_char):
                return text[:-1]
            return text

        def _create_instruction_dict(instruction=None, value=None):
            return {
                'instruction': instruction,
                'startline': lineno,
                'endline': lineno,
                'content': line,  # pylint: disable=undefined-loop-variable
                'value': value
            }

        def _clean_comment_line(line):
            line = re.sub(r'^\s*#\s*', '', line)
            line = re.sub(r'\n', '', line)
            return line

        instructions = []
        lineno = -1
        line_continuation_char = '\\'
        insnre = re.compile(r'^\s*(\S+)\s+(.*)$')  # matched group is insn
        contre = re.compile(r'^.*\\\s*$')          # line continues?
        commentre = re.compile(r'^\s*#')           # line is a comment?
        directive_possible = True
        # escape directive regex
        escape_directive_re = re.compile(r'^\s*#\s*escape\s*=\s*(\\|`)\s*$', re.I)
        # syntax directive regex
        syntax_directive_re = re.compile(r'^\s*#\s*syntax\s*=\s*(.*)\s*$', re.I)

        in_continuation = False
        current_instruction = {}

        for line in self.lines:
            lineno += 1

            if directive_possible:
                # once support for python versions before 3.8 is dropped use walrus operator
                if escape_directive_re.match(line):
                    # Do the matching twice if there is a directive to avoid doing the matching
                    # for other lines
                    match = escape_directive_re.match(line)
                    line_continuation_char = match.group(1)
                    contre = re.compile(r'^.*' + re.escape(match.group(1)) + r'\s*$')
                elif syntax_directive_re.match(line):
                    # Currently no information for the syntax directive is stored it is still
                    # necessary to detect escape directives after a syntax directive
                    pass
                else:
                    directive_possible = False

            # It is necessary to keep instructions and comment parsing separate,
            # as a multi-line instruction can be interjected with comments.
            if commentre.match(line):
                comment = _create_instruction_dict(
                    instruction=COMMENT_INSTRUCTION,
                    value=_clean_comment_line(line)
                )
                instructions.append(comment)

            else:
                if not in_continuation:
                    m = insnre.match(line)
                    if not m:
                        continue
                    current_instruction = _create_instruction_dict(
                        instruction=m.groups()[0].upper(),
                        value=_rstrip_eol(m.groups()[1], line_continuation_char)
                    )
                else:
                    current_instruction['content'] += line
                    current_instruction['endline'] = lineno

                    if current_instruction['value']:
                        current_instruction['value'] += _rstrip_eol(line, line_continuation_char)
                    else:
                        current_instruction['value'] = _rstrip_eol(line.lstrip(),
                                                                   line_continuation_char)

                in_continuation = contre.match(line)
                if not in_continuation and current_instruction:
                    instructions.append(current_instruction)

        return instructions

    @property
    def json(self):
        """
        :return: JSON formatted string with instructions & values from Dockerfile
        """
        insndescs = [{insndesc['instruction']: insndesc['value']} for insndesc in self.structure]
        return json.dumps(insndescs)

    @property
    def parent_images(self):
        """
        :return: list of parent images -- one image per each stage's FROM instruction
        """
        in_stage = False
        top_args = {}
        parents = []
        for instr in self.structure:
            if instr['instruction'] == 'ARG':
                if not in_stage:
                    key_val_list = extract_key_values(
                        env_replace=False,
                        args={}, envs={},
                        instruction_value=instr['value'])
                    for key, value in key_val_list:
                        if key in self.build_args:
                            value = self.build_args[key]
                        top_args[key] = value
            elif instr['instruction'] == 'FROM':
                in_stage = True
                image, _ = image_from(instr['value'])
                if image is not None:
                    image = WordSplitter(image, args=top_args).dequote()
                    parents.append(image)
        return parents

    @parent_images.setter
    def parent_images(self, parents):
        """
        setter for images in 'FROM' instructions.
        Images are updated per build stage with the given parents in the order they appear.
        Raises RuntimeError if a different number of parents are given than there are stages
        as that is likely to be a mistake.

        :param parents: list of image strings
        """
        parents = list(parents)
        change_instrs = []
        for instr in self.structure:
            if instr['instruction'] != 'FROM':
                continue

            old_image, stage = image_from(instr['value'])
            if old_image is None:
                continue  # broken FROM, fixing would just confuse things
            if not parents:
                raise RuntimeError("not enough parents to match build stages")

            image = parents.pop(0)
            if image != old_image:
                instr['value'] = '{0} AS {1}'.format(image, stage) if stage else image
                instr['content'] = 'FROM {0}\n'.format(instr['value'])
                change_instrs.append(instr)

        if parents:
            raise RuntimeError("trying to update too many parents for build stages")

        lines = self.lines
        for instr in reversed(change_instrs):
            lines[instr['startline']:instr['endline']+1] = [instr['content']]

        self.lines = lines

    @property
    def is_multistage(self):
        return len(self.parent_images) > 1

    @property
    def baseimage(self):
        """
        :return: base image, i.e. value of final stage FROM instruction
        """
        return (self.parent_images or [None])[-1]

    @baseimage.setter
    def baseimage(self, new_image):
        """
        change image of final stage FROM instruction
        """
        images = []
        for instr in self.structure:
            if instr['instruction'] == 'FROM':
                image, _ = image_from(instr['value'])
                if image is not None:
                    images.append(image)
        if not images:
            raise RuntimeError('No stage defined to set base image on')
        images[-1] = new_image
        self.parent_images = images

    @property
    def cmd(self):
        """
        Determine the final CMD instruction, if any, in the final build stage.
        CMDs from earlier stages are ignored.
        :return: value of final stage CMD instruction
        """
        value = None
        for insndesc in self.structure:
            if insndesc['instruction'] == 'FROM':  # new stage, reset
                value = None
            elif insndesc['instruction'] == 'CMD':
                value = insndesc['value']
        return value

    @cmd.setter
    def cmd(self, value):
        """
        setter for final 'CMD' instruction in final build stage

        """
        cmd = None
        for insndesc in self.structure:
            if insndesc['instruction'] == 'FROM':  # new stage, reset
                cmd = None
            elif insndesc['instruction'] == 'CMD':
                cmd = insndesc

        new_cmd = 'CMD ' + value
        if cmd:
            self.add_lines_at(cmd, new_cmd, replace=True)
        else:
            self.add_lines(new_cmd)

    @property
    def labels(self):
        """
        LABELs from Dockerfile
        :return: dictionary of label:value (value might be '')
        """
        return self._instruction_getter('LABEL', env_replace=self.env_replace)

    @property
    def envs(self):
        """
        ENVs from Dockerfile
        :return: dictionary of env_var_name:value (value might be '')
        """
        return self._instruction_getter('ENV', env_replace=self.env_replace)

    @property
    def args(self):
        """
        ARGs from Dockerfile
        :return: dictionary of arg_var_name:value (value might be '')
        """
        return self._instruction_getter('ARG', env_replace=self.env_replace)

    def _instruction_getter(self, name, env_replace):
        """
        Get LABEL or ENV or ARG instructions with environment replacement

        :param name: e.g. 'LABEL' or 'ENV' or 'ARG'
        :param env_replace: bool, whether to perform ENV substitution
        :return: Labels instance or Envs instance
        """
        if name not in ('LABEL', 'ENV', 'ARG'):
            raise ValueError("Unsupported instruction '{0}'".format(name))
        in_stage = False
        top_args = {}
        instructions = {}
        args = {}
        envs = {}

        for instruction_desc in self.structure:
            this_instruction = instruction_desc['instruction']
            if this_instruction == 'FROM':
                in_stage = True
                instructions.clear()
                args = {}
                envs = self.parent_env.copy()
            elif this_instruction in (name, 'ENV', 'ARG'):
                logger.debug("%s value: %r", name.lower(), instruction_desc['value'])
                key_val_list = extract_key_values(
                    env_replace=this_instruction != 'ARG' and env_replace,
                    args=args, envs=envs,
                    instruction_value=instruction_desc['value'])
                for key, value in key_val_list:
                    if this_instruction == 'ARG':
                        if in_stage:
                            if key in top_args:
                                value = top_args[key]
                            elif key in self.build_args:
                                value = self.build_args[key]
                            args[key] = value
                        else:
                            if key in self.build_args:
                                value = self.build_args[key]
                            top_args[key] = value
                    if this_instruction == name:
                        instructions[key] = value
                        logger.debug("new %s %r=%r", name.lower(), key, value)
                    if this_instruction == 'ENV':
                        envs[key] = value

        logger.debug("instructions: %r", instructions)
        if name == 'LABEL':
            return Labels(instructions, self)
        elif name == 'ENV':
            return Envs(instructions, self)
        else:
            return Args(instructions, self)

    @labels.setter
    def labels(self, labels):
        """
        Setter for LABEL instruction, i.e. sets LABELs per input param.
        :param labels: dictionary of label name & value to be set
        """
        self._instructions_setter('LABEL', labels)

    @envs.setter
    def envs(self, envs):
        """
        Setter for ENV instruction, i.e. sets ENVs per input param.
        :param envs: dictionary of env. var. name & value to be set
        """
        self._instructions_setter('ENV', envs)

    @args.setter
    def args(self, args):
        """
        Setter for ARG instruction, i.e. sets ARGs per input param.
        :param args: dictionary of arg names & values to be set
        """
        self._instructions_setter('ARG', args)

    def _instructions_setter(self, name, instructions):
        if not isinstance(instructions, dict):
            raise TypeError('instructions needs to be a dictionary {name: value}')

        if name == 'LABEL':
            existing = self.labels
        elif name == 'ENV':
            existing = self.envs
        elif name == 'ARG':
            existing = self.args
        else:
            raise ValueError("Unexpected instruction '%s'" % name)

        logger.debug("setting %s instructions: %r", name, instructions)

        to_delete = [k for k in existing if k not in instructions]
        for key in to_delete:
            logger.debug("delete %r", key)
            self._modify_instruction_label_env(name, key, None)

        to_add = dict((k, v) for (k, v) in instructions.items() if k not in existing)
        for k, v in to_add.items():
            logger.debug("add %r", k)
            self._add_instruction(name, (k, v))

        to_change = dict((k, v) for (k, v) in instructions.items()
                         if (k in existing and v != existing[k]))
        for k, v in to_change.items():
            logger.debug("modify %r", k)
            self._modify_instruction_label_env(name, k, v)

    def _modify_instruction_label(self, label_key, instr_value):
        self._modify_instruction_label_env('LABEL', label_key, instr_value)

    def _modify_instruction_env(self, env_var_key, env_var_value):
        self._modify_instruction_label_env('ENV', env_var_key, env_var_value)

    def _modify_instruction_arg(self, arg_key, arg_value):
        self._modify_instruction_label_env('ARG', arg_key, arg_value)

    def _modify_instruction_label_env(self, instruction, instr_key, instr_value):
        """
        set <INSTRUCTION> instr_key to instr_value

        :param instr_key: str, label key
        :param instr_value: str or None, new label/env value or None to remove
        """
        if instruction == 'LABEL':
            instructions = self.labels
        elif instruction == 'ENV':
            instructions = self.envs
        elif instruction == 'ARG':
            instructions = self.args
        else:
            raise ValueError("Unknown instruction '%s'" % instruction)

        if instr_key not in instructions:
            raise KeyError('%s not in %ss' % (instr_key, instruction))

        # extract target instructions from the final stage only
        candidates = []
        for insn in self.structure:
            if insn['instruction'] == 'FROM':
                candidates = []
            if insn['instruction'] == instruction:
                candidates.append(insn)

        # Find where in the file to put the changes
        content = startline = endline = None
        for candidate in candidates:
            words = list(WordSplitter(candidate['value']).split(dequote=False))

            # LABEL/ENV/ARG syntax is one of two types:
            if '=' not in words[0]:  # LABEL/ENV/ARG name value
                # Remove quotes from key name and see if it's the one
                # we're looking for.
                if WordSplitter(words[0]).dequote() == instr_key:
                    if instr_value is None:
                        # Delete this line altogether
                        content = None
                    else:
                        # Adjust label/env value
                        words[1:] = [quote(instr_value)]

                        # Now reconstruct the line
                        content = " ".join([instruction] + words) + '\n'

                    startline = candidate['startline']
                    endline = candidate['endline']
                    break
            else:  # LABEL/ENV/ARG "name"="value"
                for index, token in enumerate(words):
                    key, _ = token.split("=", 1)
                    if WordSplitter(key).dequote() == instr_key:
                        if instr_value is None:
                            # Delete this label
                            del words[index]
                        else:
                            # Adjust label/env value
                            words[index] = "{0}={1}".format(key,
                                                            quote(instr_value))

                        if len(words) == 0:
                            # We removed the last label/env, delete the whole line
                            content = None
                        else:
                            # Now reconstruct the line
                            content = " ".join([instruction] + words) + '\n'

                        startline = candidate['startline']
                        endline = candidate['endline']
                        break

        # We know the label/env we're looking for is there
        assert startline and endline

        # Re-write the Dockerfile
        lines = self.lines
        del lines[startline:endline + 1]
        if content:
            lines.insert(startline, content)
        self.lines = lines

    def _delete_instructions(self, instruction, value=None):
        """
        :param instruction: name of instruction to be deleted
        :param value: if specified, delete instruction only when it has this value
                      if instruction is LABEL then value is label name
        """
        if instruction == 'LABEL' and value:
            self._modify_instruction_label(value, None)
            return
        if instruction == 'ENV' and value:
            self._modify_instruction_env(value, None)
            return
        if instruction == 'ARG' and value:
            self._modify_instruction_arg(value, None)
            return

        lines = self.lines
        deleted = False
        for insn in reversed(self.structure):
            if insn['instruction'] == instruction:
                if value and insn['value'] != value:
                    continue
                deleted = True
                del lines[insn['startline']:insn['endline'] + 1]
        if deleted:
            self.lines = lines

    def _add_instruction(self, instruction, value):
        """
        :param instruction: instruction name to be added
        :param value: instruction value
        """
        if instruction in ('LABEL', 'ENV', 'ARG') and len(value) == 2:
            new_line = instruction + ' ' + '='.join(map(quote, value)) + '\n'
        else:
            new_line = '{0} {1}\n'.format(instruction, value)
        if new_line:
            lines = self.lines
            if not lines[len(lines) - 1].endswith('\n'):
                new_line = '\n' + new_line
            lines += new_line
            self.lines = lines

    def add_lines(self, *lines, **kwargs):
        """
        Add lines to the beginning or end of the build.
        :param lines: one or more lines to add to the content, by default at the end.
        :param all_stages: bool for whether to add in all stages for a multistage build
                           or (by default) only the last.
        :param at_start: adds at the beginning (after FROM) of the stage(s) instead of the end.
        :param skip_scratch: skip stages which use "FROM scratch"
        """
        assert len(lines) > 0
        lines = [_endline(line) for line in lines]
        all_stages = kwargs.pop('all_stages', False)
        at_start = kwargs.pop('at_start', False)
        skip_scratch = kwargs.pop('skip_scratch', False)
        assert not kwargs, "Unknown keyword argument(s): {0}".format(list(kwargs))

        froms = [
            instr for instr in self.structure
            if instr['instruction'] == 'FROM'
        ] or [{'endline': -1}]  # no FROM? fake one before the beginning
        if not all_stages:  # only modify the last
            froms = [froms[-1]]

        df_lines = self.lines
        # make sure last line has a newline if lines are to be appended
        if df_lines and not at_start:
            df_lines[-1] = _endline(df_lines[-1])

        # iterate through the stages in reverse order
        # so adding lines doesn't invalidate line numbers from structure dicts.
        # first add a bogus instruction to represent EOF in our iteration.
        froms.append({'startline': len(df_lines) + 1})
        for stage in range(len(froms)-2, -1, -1):  # e.g. 0 for single or 2, 1, 0 for 3 stages
            start, finish = froms[stage], froms[stage+1]
            linenum = start['endline'] + 1 if at_start else finish['startline']
            image, _ = image_from(froms[stage].get('value') or '')
            if skip_scratch and image == 'scratch':
                continue
            df_lines[linenum:linenum] = lines

        self.lines = df_lines

    def add_lines_at(self, anchor, *lines, **kwargs):
        """
        Add lines at a specific location in the file.
        :param anchor: structure_dict|line_str|line_num a reference to where adds should occur
        :param lines: one or more lines to add to the content
        :param replace: if True -- replace the anchor
        :param after: if True -- insert after the anchor (conflicts with "replace")
        """
        assert len(lines) > 0
        replace = kwargs.pop('replace', False)
        after = kwargs.pop('after', False)
        assert not (after and replace)
        assert not kwargs, "Unknown keyword argument(s): {0}".format(list(kwargs))

        # find the line number for the insertion
        df_lines = self.lines
        if isinstance(anchor, int):  # line number, just validate
            assert 0 <= anchor < len(df_lines)
            if replace:
                del df_lines[anchor]
        elif isinstance(anchor, dict):  # structure
            assert anchor in self.structure, "Current structure does not match: {0}".format(anchor)
            if replace:
                df_lines[anchor['startline']:anchor['endline'] + 1] = []
            if after:
                anchor = anchor['endline']
            else:
                anchor = anchor['startline']
        elif isinstance(anchor, str):  # line contents
            matches = [index for index, text in enumerate(df_lines) if text == anchor]
            if not matches:
                raise RuntimeError("Cannot find line in the build file:\n" + anchor)
            anchor = matches[-1]
            if replace:
                del df_lines[anchor]
        else:
            raise RuntimeError("Unknown anchor type {0}".format(anchor))

        if after:
            # ensure there's a newline on final line
            df_lines[anchor] = _endline(df_lines[anchor])
            anchor += 1

        df_lines[anchor:anchor] = [_endline(line) for line in lines]
        self.lines = df_lines

    @property
    def context_structure(self):
        """
        :return: list of Context objects
            (Contains info about build arguments, labels, and environment variables for each line.)
        """
        in_stage = False
        top_args = {}
        instructions = []
        last_context = Context()
        for instr in self.structure:
            instruction_type = instr['instruction']
            if instruction_type == "FROM":  # reset per stage
                in_stage = True
                last_context = Context(envs=dict(self.parent_env))

            context = Context(args=dict(last_context.args),
                              envs=dict(last_context.envs),
                              labels=dict(last_context.labels))

            if instruction_type in ('ARG', 'ENV', 'LABEL'):
                values = get_key_val_dictionary(
                    instruction_value=instr['value'],
                    env_replace=instruction_type != 'ARG' and self.env_replace,
                    args=last_context.args,
                    envs=last_context.envs)
                if instruction_type == 'ARG' and self.env_replace:
                    if in_stage:
                        for key in list(values.keys()):
                            if key in top_args:
                                values[key] = top_args[key]
                            elif key in self.build_args:
                                values[key] = self.build_args[key]
                    else:
                        for key, value in list(values.items()):
                            if key in self.build_args:
                                value = self.build_args[key]
                            top_args[key] = value
                            values[key] = value
                context.set_line_value(context_type=instruction_type, value=values)

            instructions.append(context)
            last_context = context
        return instructions


def image_from(from_value):
    """
    :param from_value: string like "image:tag" or "image:tag AS name"
    :return: tuple of the image and stage name, e.g. ("image:tag", None)
    """
    regex = re.compile(r"""(?xi)        # readable, case-insensitive regex
        \s*                             # ignore leading whitespace
        (?P<platform> --platform=\S+)?  # optional platform parameter
        \s*                             # more whitespaces
        (?P<image> \S+ )                # image and optional tag
        (?:                             # optional "AS name" clause for stage
            \s+ AS \s+
            (?P<name> \S+ )
        )?
        """)
    match = re.match(regex, from_value)
    return match.group('image', 'name') if match else (None, None)


def _endline(line):
    """
    Make sure the line ends with a single newline.
    Since trailing whitespace has no significance, remove it.
    """
    return line.rstrip() + '\n'
